# plik

File system in user space ([FUSE](https://en.wikipedia.org/wiki/Filesystem_in_Userspace)) with compression and deduplication.
Written in a single C file (1K LOC). Uses [OpenSSL](https://www.openssl.org/) for hashing, [zlib](https://www.zlib.net/) for compression and [SQLite](https://www.sqlite.org/) for data storage.

Can be even used as a permanent storage, or as a [ramdrive](https://en.wikipedia.org/wiki/RAM_drive) (with ":memory:" database).
Should be well suited for storing rather small, well compressible and possibly duplicate files (like a number of source trees, fetched from different sources, containing same submodules, etc).

Purely research project. Not for any other type of use.

## How it works

There are two tables created in the database. The "t\_entry" table represents file/folder hierarchy, and the "t\_block" table represents binary blobs. When the file is opened, its blob gets fetched and decompressed into cache. The following read/write operations are done to the cached byte array. When the file is closed, it is checksummed, compressed and stored into the database. If there is already a blob with the same checksum, the blob is not stored, and the file is linked to the already existing blob.

## Things to know about

1. FUSE "high level" interface partially (but enough) implemented.
2. A lot of file attributes are not handled (but enough for general usage).
3. There is no automatic testing, only selective, handmade and occasional one.
4. There was no SCM used for the development.
5. Some memory management tricks were invented (or reinvented).
6. All the build/test tasks are made with a single Makefile.
7. Binary blobs, that are dereferenced, are not removed right away (though removed on mount), and the database is not compacted. That'd be a heavy operation. And dereferenced blocks have a chance to be rereferenced occasionally.
8. All the debug messages and comments are in Russian. That's how life turns out.
9. A lot of other things, that I don't remember about at the moment.

## Download

- **Source** tarball [plik.tar.bz2](plik.tar.bz2)
- **Bazaar repo** bzr clone [https://tycho.sytes.net/plik/](https://tycho.sytes.net/plik/)
- **Direct** file links [main.c](main.c) [Makefile](Makefile) [LICENSE](LICENSE)

## Build

1. Install development dependencies, depending on your OS: fuse, zlib, sqlite, openssl.
2. Run *make plik* (and see Makefile for when can be built).

## Use

- To obtain full info on command line options:

> ./plik -h

- To mount a "ramdrive" into "test\_folder" and continue in the background:

> ./plik test\_folder -o db\_name=:memory:

- To mount a permanent test.db into "test\_folder", continue in foreground with full debug:

> ./plik test\_folder -f -d -o db\_name=test.db

There are other options, that are common, and processed by FUSE library.

## Feedback

EMail: [plik@tycho.sytes.net](mailto:plik@tycho.sytes.net)

All the feedback, suggestions, ideas, contributions are welcome.

Though nothing is promised.

## Теперь по-русски

Собственно, все образованные люди - нет необходимости пересказывать всё выдавленное выше.

### История вопроса

Уже довольно давно на свете существует [ZFS](https://ru.wikipedia.org/wiki/ZFS). И всем хорош, и сжатие, и дедубликация, и снимки, и бэкапы, и неразрушительное управление томами и всякое. Но одновременно, почти нигде толком не используется. А почему. Во-первых, все боятся [Ларри Эллисона](https://ru.wikipedia.org/wiki/%D0%AD%D0%BB%D0%BB%D0%B8%D1%81%D0%BE%D0%BD,_%D0%9B%D0%B0%D1%80%D1%80%D0%B8). Он мужчина серьёзный и у него много адвокатов. И если он придёт к вам и начнёт применять пруденцию, останетесь без штанов. Не то чтобы окончательно правильная свободная лицензия - все это понимают. Во-вторых, и тут я согласен с [Тео де Раадтом](https://ru.wikipedia.org/wiki/%D0%A0%D0%B0%D0%B0%D0%B4%D1%82,_%D0%A2%D0%B5%D0%BE_%D0%B4%D0%B5), смешивание уровней абстракции и нарушение общей логики и изоляции - объединение томов и файловых систем - это неправильно. Опять же, какие альтернативы. [BTRFS](https://ru.wikipedia.org/wiki/Btrfs), говорят, не готов (кроме этого, [эти люди](https://www.linux.org.ru/news/kernel/17799783) из этого вашего [Линукса](https://ru.wikipedia.org/wiki/Linux) недавно [ReiserFS](https://ru.wikipedia.org/wiki/ReiserFS) отпилили. А ReiserFS был охренителен). Поэтому у нас будет свой ZFS, корявый и недоделанный, в юзерспейсе.

### История реализации

Первая версия, ясное дело, была написана на [Qt](https://www.qt.io/). Но выяснилось, что [Гоблинтеч](https://ru.wikipedia.org/wiki/The_Qt_Company) мало того, что сделал QtSql не потокобезопасным (как оказалось, обвесить мьютексом недостаточно), но требующим, чтобы обращение к БД было только из главной нити. FUSE многопоточен. Можно конечно толкать в однопоточном режиме (c "-s") и будет работать. Но это не спортивно.

Далее появилась мысль взять другую библиотеку. Слышал про [Poco](https://pocoproject.org/). Почитал документацию (ту, что нашел). Перебор. Много неизведанного.

[Boost](https://www.boost.org/) или [стандартная библиотека C++](https://en.cppreference.com/w/cpp/standard_library)? Пускай [Бьёрн Страуструп](https://ru.wikipedia.org/wiki/%D0%A1%D1%82%D1%80%D0%B0%D1%83%D1%81%D1%82%D1%80%D1%83%D0%BF,_%D0%91%D1%8C%D1%91%D1%80%D0%BD) на этом пишет. Я тут недавно видел последний стандарт C++. Кино и немцы. С каждой итерацией новый язык. Ничего похожего на то, что было 20 лет назад с этим же названием. Есть мнение, что выпускаемые стандарты C++ нужны главным образом "ребятам из Яндекса", у которых нет других оснований покататься по миру и потом похвастаться на [Юттэп](https://www.youtube.com/). Могли бы просто кататься. Но нет, извратили язык дальше некуда. И хвастаются. Видел тут видео, где собрались целое помещение и угадывают, что делает программа из 4-х строк. И долго не могут угадать. Я бы вот не хвастался.

Назад к корням. Хотелось бы всяких структур данных и функций по их управлению. Посмотрел на [Glib](https://ru.wikipedia.org/wiki/GLib) и на [EFL](https://www.enlightenment.org/about-efl.md). С другой стороны, с документацией тяжеловато. Такой документации, как в Qt нет нигде. Максималистическое субъективное заявление.

Поэтому пускай будет чистый Си с частными библиотеками для специфического (а не общего) назначения. Опять же [valgrind](https://ru.wikipedia.org/wiki/Valgrind) и [gprof](https://en.wikipedia.org/wiki/Gprof) будут бойчее работать.

Вся реализация в одном [main.c](main.c) (подход "амальгамирование" принятый в [JUCE](https://ru.wikipedia.org/wiki/Juce)). Кто теряется в 1K LOC, у того низкий [ICQ](https://ru.wikipedia.org/wiki/ICQ), и тот нам не товарищ. А про то, что такое бывает и у нас, мы не признаемся, чтобы нас не обвинили в низком ICQ.

### История названия

Название взято из [Гаду-Гаду](https://gadu-gadu.pl/), лучшего на сегодняшний день мессенджера с лучшим названием и лучшими смайликами из всех, что есть на свете.

### Сама реализация

В Си нет исключений. Или есть, но мы о них не знаем. И одновременно нет возможности вернуть из функции анонимный кортеж, как в [Perl 5](https://www.perl.org/) (лучшем языке программирования на свете). Поэтому, функции почти везде сконструированы как в [классическом MacOS](https://en.wikipedia.org/wiki/Classic_Mac_OS) (и много ещё где) - набор входных и выходных параметров, в качестве возвращаемого значения bool успешности. Проверка на возврат false и возврат false - как замена исключений. Там, где false "порождается" - вывод сообщения об ошибке в консоль.

Ещё измышления. Если переменная на стеке, она вылетит с выходом из области видимости. Объект класса будет грохнут деструктором, который будет вызван в эпилоге функции компилятором C++. Для Си так не получится. Переменную на стеке он выбросит, а вот "деструктор" вызван неявно не будет. С указателями вообще беда - всё что в куче, останется в куче. Нужно всегда явно удалять то, что сам создал. Принимая во внимания ветвления и возвраты из середины функции, отслеживание созданного становится хитрой задачей. В [Objective-C](https://ru.wikipedia.org/wiki/Objective-C) есть такая штука как [NSAutoreleasePool](https://developer.apple.com/documentation/foundation/nsautoreleasepool). В [go](https://go.dev/) есть конструкции типа [defer file.Close()](https://go.dev/tour/flowcontrol/12). То есть, можно создать объект и сразу его отметить, как требующий сноса по завершении обработки, вне зависимости от того, что будет происходить дальше. В Си такого нет. Сконструируем. Будет pool, куда мы будем скучивать все промежуточные структуры (с указателем на функцию сноса), который будет для нас удалять весь мусор в конце "обработки события". Может быть человечество такое сделало и без меня. Я тоже сделал. Valgrind подтверждает корректность. Появляется необходимость расслоить непосредственно функцию, и наружную обёртку, создающую и сносящую пул. Ну что делать, пока так.

У пула нет retain'а, то есть нельзя удержать то, что уже предназначено у сносу, только перекопировать. Поэтому тут и там есть "лишнее" копирование, но не много.

Разбор аргументов в стиле FUSE тоже субъидеален на мой перфекционистский взгляд. Это кроме того, что как добавил поддержку "разбора аргументов", то тут же появилась утечка памяти. Не солидно. Даже всё настроение продолжать пропало, пока не придумаю, как устранить утечку. Переделывать полностью парсинг опций командной строки и фальсификацию оных отдельно для FUSE пока делать лениво. Поэтому остановимся пока на том, что есть.

Эта штука в целом может вообще к стати не работать. Но я бы попробовал. Самокритика и оптимизм.

### Вместо заключения

Почтовый адрес для сбора обратной связи написан чуть выше (тут квест, ага). Конструктивным замечаниям, предложениям по развитию проекта буду рад.

*Д.С.*
